/*
 * This file is part of Locutil2.
 * 
 * Copyright (c) 2007 Thomas King <king@informatik.uni-mannheim.de>,
 * University of Mannheim, Germany
 * 
 * All rights reserved.
 *
 * Loclib is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * Loclib is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with Loclib; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
 */

package org.pi4.locutil2;

import java.util.ArrayList;
import java.util.Vector;

import org.pi4.locutil2.GeoPosition;
import org.pi4.locutil2.Random;
import org.pi4.locutil2.Statistics;

import junit.framework.TestCase;

/**
 * Test cases for org.pi4.loceva.util.Statistics.
 * 
 * @author abiskop
 */
public class StatisticsTest extends TestCase {
	
	private static final double PRECISION = 0.000000001;
	private static final int NUMBER_OF_TEST_VALUES = 3;
	private static final int NUMBER_OF_RANDOM_TESTS = 100;
	
	private static double[][] values;
	private static double[] avgExpected;
	private static double[] medianExpected;
	private static double[] stdDevExpected;
	
	public void setUp() {
	    values = new double[NUMBER_OF_TEST_VALUES][];
	    avgExpected = new double[NUMBER_OF_TEST_VALUES];
	    medianExpected = new double[NUMBER_OF_TEST_VALUES];
	    stdDevExpected = new double[NUMBER_OF_TEST_VALUES];
	    
	    values[0] = new double[] {
	            4, 2, 3, 4, 5, 6
	    };
	    avgExpected[0] = 4;
	    medianExpected[0] = 4;
	    stdDevExpected[0] = Math.sqrt(2);
	    
	    values[1] = new double[] {
	            1.5, 2.5, 3.5, 4.5, 5.5
	    };
	    avgExpected[1] = 3.5;
	    medianExpected[1] = 3.5;
	    stdDevExpected[1] = Math.sqrt(2.5);
	    
	    values[2] = new double[] {
	            -1.25, 1.25
	    };
	    avgExpected[2] = 0;
	    medianExpected[2] = 0;
	    stdDevExpected[2] = Math.sqrt(3.125);
	}
	
	public void tearDown() {
	}
	
    /**
     * Test method for Statistics.avg(double[]).
     */
	public void testAvgDouble() {
	    for (int i = 0; i < NUMBER_OF_TEST_VALUES; i++) {
	        assertEquals(avgExpected[i], Statistics.avg(values[i]), PRECISION);
	    }
	}
	
    /**
     * Test method for Statistics.avg(ArrayList<? extends Number>).
     */
	public void testAvgArrayList() {
		ArrayList<Long> al = new ArrayList<Long>();
		al.add(Long.MAX_VALUE);
		al.add(Long.MAX_VALUE);
		al.add(Long.MAX_VALUE);
		al.add(Long.MAX_VALUE);
		assertEquals("Long MAX_VALUE: ", (double) Long.MAX_VALUE, Statistics.avg(al), Math.pow(10, 4));
		
		al.clear();
		al.add(Long.MIN_VALUE);
		al.add(Long.MIN_VALUE);
		al.add(Long.MIN_VALUE);
		al.add(Long.MIN_VALUE);
		assertEquals("Long MIN_VALUE: ", (double) Long.MIN_VALUE, Statistics.avg(al));
		
		ArrayList<Double> ad = new ArrayList<Double>();
		ad.add(Double.MAX_VALUE);
		ad.add(Double.MAX_VALUE);
		assertTrue("Double MAX_VALUE: ", Double.isInfinite(Statistics.avg(ad)));
		
		ad.clear();
		ad.add(Double.MIN_VALUE);
		ad.add(Double.MIN_VALUE);
		ad.add(Double.MIN_VALUE);
		ad.add(Double.MIN_VALUE);
		assertEquals("Double MIN_VALUE: ", Double.MIN_VALUE, Statistics.avg(ad));
	}
	
	public void testAvgGeoPosition() {
		GeoPosition one = new GeoPosition(1.5, 1.5, 1.5, 50.5);
		GeoPosition two = new GeoPosition(1.0, 1.0, 1.0, 20.0);
		Vector<GeoPosition> list = new Vector<GeoPosition>();
		list.add(one);
		list.add(two);
		assertEquals(new GeoPosition(1.25, 1.25, 1.25, 35.25), Statistics.avg(list));
	}
	
	public void testMax_a() {
		ArrayList<Integer> array = new ArrayList<Integer>();
		array.add(1);
		array.add(2);
		array.add(3);
		assertEquals(3.0, Statistics.max(array));
	}
	
	public void testMax_b() {
		ArrayList<Integer> array = new ArrayList<Integer>();
		array.add(3);
		array.add(2);
		array.add(1);
		assertEquals(3.0, Statistics.max(array));
	}
	
	public void testMin_a() {
		ArrayList<Integer> array = new ArrayList<Integer>();
		array.add(1);
		array.add(2);
		array.add(3);
		assertEquals(1.0 , Statistics.min(array));
	}
	
	public void testMin_b() {
		ArrayList<Integer> array = new ArrayList<Integer>();
		array.add(3);
		array.add(2);
		array.add(1);
		assertEquals(1.0, Statistics.min(array));
	}
	
    /**
     * Test method for Statistics.median(double[]).
     */
	public void testMedian() {
	    for (int i = 0; i < NUMBER_OF_TEST_VALUES; i++) {
	        assertEquals(medianExpected[i], Statistics.median(values[i]), PRECISION);
	    }
	}
	
    /**
     * Test method for Statistics.stdDev(double[]).
     */
	public void testStdDev() {
	    for (int i = 0; i < NUMBER_OF_TEST_VALUES; i++) {
	        assertEquals(stdDevExpected[i], Statistics.stdDev(values[i]), PRECISION);
	    }
	}
	
    /**
     * Test method for Statistics.spearmanCoefficient(double[], double[]).
     */
	public void testSpearman_a() {
		double[] a = new double[]{2,1,3};
		assertEquals(1.0, Statistics.spearmanCoefficient(a, a), 0.0);
	}
	
    /**
     * Tests Statistics.spearmanCoefficient(double[], double[]) using randomized values.
     */
	public void testSpearman_random() {
		for (int i = 0; i < NUMBER_OF_RANDOM_TESTS; i++) {
			int len = Random.nextInt(Short.MAX_VALUE - 2) + 2;
			double[] data1 = new double[len];
			double[] data2 = new double[data1.length];
			for (int j = 0; j < len; j++) {
				data1[j] = Random.nextInt();
				data2[j] = Random.nextInt();
			}
			double rs = Statistics.spearmanCoefficient(data1, data2);
			assertTrue("rs out of range: " + rs + " ("+len+")", (rs >= -1.0) && (rs <= 1.0));
			assertEquals(1.0, Statistics.spearmanCoefficient(data1, data1), 0.0);
			assertEquals(1.0, Statistics.spearmanCoefficient(data2, data2), 0.0);
		}
	}
	
    /**
     * Test method for Statistics.approximationCumulativeDistributionFunction(..).
     */
	public void testApproximationCumulativeDistributionFunction() {
		assertTrue(Statistics.approximationCumulativeDistributionFunction(0, 0, 1) < 0.5001);
		assertTrue(Statistics.approximationCumulativeDistributionFunction(0, 0, 1) > 0.4968);
		
		assertTrue(Statistics.approximationCumulativeDistributionFunction(2.51, 0, 1) < 0.9940);
		assertTrue(Statistics.approximationCumulativeDistributionFunction(2.51, 0, 1) > 0.9930);

		assertTrue(Statistics.approximationCumulativeDistributionFunction(3.49, 0, 1) < 0.9999);
		assertTrue(Statistics.approximationCumulativeDistributionFunction(3.49, 0, 1) > 0.9992);
		
		assertEquals(0.0, Statistics.approximationCumulativeDistributionFunction(0.5, 1, 0));
		assertEquals(1.0, Statistics.approximationCumulativeDistributionFunction(1.5, 1, 0));
		
		assertTrue(Statistics.approximationCumulativeDistributionFunction(7, 5.0, 2.0) - Statistics.approximationCumulativeDistributionFunction(3, 5.0, 2.0) > 0.68100);
		assertTrue(Statistics.approximationCumulativeDistributionFunction(7, 5.0, 2.0) - Statistics.approximationCumulativeDistributionFunction(3, 5.0, 2.0) < 0.68300);
	}
}
